; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         ns-gdb.el
; RCS:          /usr/local/lib/cvs/emacs-19/lisp/ns-gdb.el,v 1.1.2.4 1994/05/25 01:32:44 cedman Exp
; Description:  A mode to support interaction between Emacs and NeXT gdb
; Author:       Carl Edman
; Created:      Sun May 15 15:08:45 1994
; Modified:     Wed May 18 20:59:01 1994 (Carl Edman) cedman@capitalist.princeton.edu
; Language:     Emacs-Lisp
; Package:      N/A
; Status:       Experimental (Do Not Distribute)
;
; (C) Copyright 1994, but otherwise this file is perfect freeware.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ns-gdb-mode nil
  "Non-nil when in NS gdb mode.")
(make-variable-buffer-local 'ns-gdb-mode)
(defvar ns-gdb-mode-map nil
  "Keymap used by NS gdb mode.")

(defun ns-gdb-mode (arg)
  "Toggle NS gdb mode.\n\
With argument ARG, turn mode on iff ARG is positive.\n\
\\{ns-gdb-mode-map}"
  (interactive "P")
  (setq ns-gdb-mode
        (if (null arg) (not ns-gdb-mode)
          (> (prefix-numeric-value arg) 0)))
  (force-mode-line-update))

(if ns-gdb-mode-map
    nil
  (setq ns-gdb-mode-map (make-sparse-keymap))
  (define-key ns-gdb-mode-map "\C-cn" 'ns-gdb-next)
  (define-key ns-gdb-mode-map "\C-cs" 'ns-gdb-step)
  (define-key ns-gdb-mode-map "\C-cc" 'ns-gdb-continue)
  (define-key ns-gdb-mode-map "\C-cf" 'ns-gdb-finish)
  (define-key ns-gdb-mode-map "\C-cq" 'ns-gdb-quit)
  (define-key ns-gdb-mode-map "\C-cr" 'ns-gdb-run)
  
  (define-key ns-gdb-mode-map "\C-cD" 'ns-gdb-down)
  (define-key ns-gdb-mode-map "\C-cU" 'ns-gdb-up)
  (define-key ns-gdb-mode-map "\C-cF" 'ns-gdb-frame)
  (define-key ns-gdb-mode-map "\C-cS" 'ns-gdb-select-frame)
  
  (define-key ns-gdb-mode-map "\C-cb" 'ns-gdb-break-here)
  (define-key ns-gdb-mode-map "\C-cu" 'ns-gdb-until-here)
  (define-key ns-gdb-mode-map "\C-cp" 'ns-gdb-print-here)
  (define-key ns-gdb-mode-map "\C-cP" 'ns-gdb-print-indirect-here)
  (define-key ns-gdb-mode-map "\C-c\C-g" 'ns-gdb-interrupt)
  (define-key ns-gdb-mode-map "\C-c\C-c" 'ns-gdb-mode))

(defun ns-gdb-break-here ()
  "Put a break point on the current line in the current file."
  (interactive)
  (ns-gdb-break (file-name-nondirectory (buffer-file-name))
                (1+ (count-lines (point-min) (point)))))

(defun ns-gdb-until-here ()
  "Run until reaching the current line in the current file."
  (interactive)
  (ns-gdb-until (file-name-nondirectory (buffer-file-name))
                (1+ (count-lines (point-min) (point)))))

(defun ns-gdb-print-here ()
  "Print the value of the word around point."
  (interactive)
  (ns-gdb-print (current-word)))

(defun ns-gdb-print-indirect-here ()
  "Print the value pointed by the word around the point."
  (interactive)
  (ns-gdb-print-indirect (current-word)))

(or (assq 'ns-gdb-mode minor-mode-alist)
    (setq minor-mode-alist (cons '(ns-gdb-mode (" NS gdb:" ns-gdb-status))
                                 minor-mode-alist)))
(or (assq 'ns-gdb-mode minor-mode-map-alist)
    (setq minor-mode-map-alist (cons (cons 'ns-gdb-mode ns-gdb-mode-map)
                                     minor-mode-map-alist)))
