" Vim syntax file
" Language:	shell (sh)
" Maintainer:	Lennart Schultz
" Last change:	1997 April 20

" Remove any old syntax stuff hanging around
syntax clear

syntax region shSinglequote start=+'+  skip=+\\'+  end=+'+  
" A bunch of useful sh keywords
syntax keyword shFunction	    function
syntax keyword shStatement    bg break cd chdir continue echo eval exec 
syntax keyword shStatement    exit export fg getopts hash jobs kill 
syntax keyword shStatement    newgrp pwd read readonly return set shift
syntax keyword shStatement    stop suspend test times trap type ulimit
syntax keyword shStatement    umask unset wait
syntax keyword shConditional  if else esac case then elif fi in
syntax keyword shRepeat       while for do done
syntax match   shSource       "^[ 	]*[.][ 	]\+\|[ 	]\+[.][ 	]\+" 
syntax region  shColon        start="^[ 	]*:" end="$\|#" contains ALLBUT shFunction

syntax match  shIdentifier      "\<[a-zA-Z_][a-zA-Z0-9_]*\>"
" Following is worth to notice: command substitution, file redirection and functions (so these features turns red)
syntax match   shFunctionName "[a-zA-Z_][a-zA-Z0-9_]*[ 	]*()"
syntax match   shCurlyList    "[{][ ]\=\|\;\=\}"
syntax region  shCommandSub   start="`" skip="\\`" end="`" contains ALLBUT shFunction
syntax match   shRedirStdOut2  ">>-\="
syntax match   shRedirStdOut1  "[0-9]\=>\(\(&[-0-9]\)\=\|[^>]E-1\)"
syntax match   shRedirStdIn2   "<<-\="
syntax match   shRedirStdIn1   "[0-9]\=<\(\(&[-0-9]\)\=\|[^<]E-1\)"

syntax keyword shTodo contained TODO

syntax keyword shShellVariables		HOME PATH CDPATH SHELL LC_TYPE LC_MESSAGE
syntax keyword shShellVariables		MAIL MAILCHECK PS1 PS2 IFS SHACCT 
syntax match  shSpecialShellVar         "\$[-#@*$?!0-9]" 
syntax keyword shSetVariables		ignoreeof noclobber
syntax region shDerefOpr	start="\${" end="}" contains shShellVariables shIdentifier
syntax match  shDerefIdentifier      "\$\<[a-zA-Z_][a-zA-Z0-9_]*\>"
syntax match   shOperator         "[][&;|)(]"

" String and Character contstants
" Highlight special characters (those which have a backslash) differently
syntax match   shSpecial           contained "\\[0-9][0-9][0-9]\|\\[abcfnrtv\\]"


syntax match  shNumber          "-\=\<[0-9]\+\>"
syntax match  shComment         "#.*$" contains shNumber shTodo 


syntax match shTestOpr	"-\<[oeaznlg][tfqet]\=\>\|!\==\|-\<[b-gkLprsStuwjxOG]\>"
"syntax region shTest           start="\[" skip="\\$" end="\]" contains shString shTestOpr shDerefIdentifier shDerefOpr
syntax region  shString            start=+"+  skip=+\\"+  end=+"+  contains shSpecial shOperator shDerefIdentifier shDerefOpr shSpecialShellVar shSinglequote shCommandSub

if !exists("did_sh_syntax_inits")
  let did_sh_syntax_inits = 1
  " The default methods for highlighting.  Can be overridden later
  highlight link shSinglequote		shString
  highlight link shConditional		shStatement
  highlight link shRepeat		shStatement
  highlight link shSource		shStatement
  highlight link shColon		shStatement
  highlight link shFunctionName		shFunction
  highlight link shCurlyList		shFunction
  highlight link shCommandSub		shFunction
  highlight link shRedirStdOut1		shFunction
  highlight link shRedirStdOut2		shFunction
  highlight link shRedirStdIn1		shFunction
  highlight link shRedirStdIn2		shFunction
  highlight link shSetVariables		shShellVariables
  highlight link shSpecialShellVar	shShellVariables
  highlight link shTestOpr		shOperator
  highlight link shDerefOpr		shSpecial
  highlight link shDerefIdentifier	shSpecial
  highlight link shStatement		Statement
  highlight link shNumber		Number
  highlight link shString		String
  highlight link shComment		Comment
  highlight link shSpecial		Special
  highlight link shTodo			Todo
  highlight link shIdentifier		Identifier
  highlight shOperator		term=underline ctermfg=6 guifg=Purple gui=bold
  highlight shShellVariables	term=underline ctermfg=2 guifg=SeaGreen gui=bold
  highlight shFunction		term=bold ctermbg=1 guifg=Red
endif
