" Vim syntax support file
" Maintainer:	Bram Moolenaar <mool@oce.nl>
" Last change:	1997 April 20

" This is the startup file for syntax highlighting.
" It sets the default highlighting methods, and installs autocommands for all
" the available syntax files.

if exists("has_syntax")

  " The default methods for highlighting.  Can be overridden later.
  " There should be only six of these, because many terminals can only use
  " six different colors (plus black and white).  It doesn't look nice with
  " too many colors too.
  highlight Statement	start=t_md stop=t_me cterm=bold ctermfg=5 gui=bold guifg=Brown
  highlight PreCondit	term=underline ctermfg=1 guifg=Purple
  highlight Type	term=underline ctermfg=2 guifg=SeaGreen gui=bold
  highlight String	term=italic ctermfg=2 guifg=Magenta gui=underline
  highlight Comment	term=bold ctermfg=4 guifg=Blue
  highlight Special	term=bold ctermfg=6 guifg=SlateBlue gui=underline
  highlight Identifier	term=underline ctermfg=2 guifg=DarkCyan

  " These two change the background
  highlight Error	term=reverse ctermbg=1 guibg=Red
  highlight Todo	term=standout ctermbg=3 guifg=Blue guibg=Yellow

  " Common groups that don't have default highlighting, but you can add them
  " easily.  TODO: make the list longer.
  highlight link Number		String
  highlight link Character	String
  highlight link Constant	String
  highlight link Function	Statement
  highlight link Label		Statement
  highlight link Operator	Statement
  highlight link Macro		PreCondit
  highlight link Include	PreCondit

  " Ada (83, 9X, 95)
  au BufNewFile,BufReadPost *.adb,*.ads		source $VIM/syntax/ada.vim

  " Assembly (GNU)
  au BufNewFile,BufReadPost *.asm,*.s		source $VIM/syntax/asm.vim

  " Batch files for MSDOS
  au BufNewFile,BufReadPost *.bat,*.sys		source $VIM/syntax/dosbatch.vim

  " C
  au BufNewFile,BufReadPost *.c			source $VIM/syntax/c.vim

  " C++
  au BufNewFile,BufReadPost *.cpp,*.cc,*.h,*.cxx,*.c++,*.C,*.hh	source $VIM/syntax/cpp.vim

  " Objective C
  au BufNewFile,BufReadPost *.m,*.h		source $VIM/syntax/objc.vim

  " Fortran
  au BufNewFile,BufReadPost *.f,*.for,*.fpp	source $VIM/syntax/fortran.vim

  " Help files for Vim
  au BufNewFile,BufReadPost */vim*/doc/*.txt	source $VIM/syntax/help.vim

  " HTML
  au BufNewFile,BufReadPost *.html,*.htm	source $VIM/syntax/html.vim

  " Java
  au BufNewFile,BufReadPost *.java		source $VIM/syntax/java.vim

  " Makefiles
  au BufNewFile,BufReadPost [mM]akefile*	source $VIM/syntax/make.vim

  " Pascal
  au BufNewFile,BufReadPost *.p,*.pas		source $VIM/syntax/pascal.vim

  " Perl
  au BufNewFile,BufReadPost *.pl,*.pm		source $VIM/syntax/perl.vim

  " PostScript
  au BufNewFile,BufReadPost *.ps,*.eps		source $VIM/syntax/postscr.vim

  " Shell scripts (until there is a csh.vim)
  au BufNewFile,BufReadPost .profile,.login,.cshrc,.bashrc	source $VIM/syntax/sh.vim

  " SQL
  au BufNewFile,BufReadPost *.sql		source $VIM/syntax/sql.vim

  " Tex
  au BufNewFile,BufReadPost *.tex		source $VIM/syntax/tex.vim

  " Verilog HDL (just a guess!)
  au BufNewFile,BufReadPost *.hdl		source $VIM/syntax/verilog.vim

  " Vim scripts
  au BufNewFile,BufReadPost *.*vimrc,*.vim,.exrc,_*vimrc,_exrc source $VIM/syntax/vim.vim

  " Xmath
  au BufNewFile,BufReadPost *.ms,*.msc,*.msf	source $VIM/syntax/xmath.vim

  " Various scripts, without a specific extension
  au BufNewFile,BufReadPost *			source $VIM/syntax/scripts.vim

  if exists("has_gui")
    nmenu Syntax.OFF		:syntax clear<CR>
    nmenu Syntax.Ada		:source $VIM/syntax/ada.vim<CR>
    nmenu Syntax.assembly	:source $VIM/syntax/asm.vim<CR>
    nmenu Syntax.MS-DOS\ \.bat	:source $VIM/syntax/dosbatch.vim<CR>
    nmenu Syntax.C		:source $VIM/syntax/c.vim<CR>
    nmenu Syntax.C++		:source $VIM/syntax/cpp.vim<CR>
	nmenu Syntax.ObjC		:source $VIM/syntax/objc.vim<CR>
    nmenu Syntax.Fortran	:source $VIM/syntax/fortran.vim<CR>
    nmenu Syntax.HTML		:source $VIM/syntax/html.vim<CR>
    nmenu Syntax.Java		:source $VIM/syntax/java.vim<CR>
    nmenu Syntax.Makefile	:source $VIM/syntax/make.vim<CR>
    nmenu Syntax.Matlab		:source $VIM/syntax/matlab.vim<CR>
    nmenu Syntax.Pascal		:source $VIM/syntax/pascal.vim<CR>
    nmenu Syntax.Perl		:source $VIM/syntax/perl.vim<CR>
    nmenu Syntax.PostScript	:source $VIM/syntax/postscr.vim<CR>
    nmenu Syntax.Shell\ script	:source $VIM/syntax/sh.vim<CR>
    nmenu Syntax.SQL		:source $VIM/syntax/sql.vim<CR>
    nmenu Syntax.Tex		:source $VIM/syntax/tex.vim<CR>
    nmenu Syntax.Verilog\ HDL	:source $VIM/syntax/verilog.vim<CR>
    nmenu Syntax.Vim\ help	:source $VIM/syntax/help.vim<CR>
    nmenu Syntax.Vim\ script	:source $VIM/syntax/vim.vim<CR>
    nmenu Syntax.Xmath		:source $VIM/syntax/xmath.vim<CR>
  endif

endif
