#
# Makefile for Vim on MiNT				vim:ts=8:sw=8:tw=78
#
# This is a modified version of the original vim 4.0 Makefile. It is
# configured to accompany the MiNT distribution of Vim.
# The GUI, Win32, Amiga, MS-Dos, and cproto stuff has been deleted because
# the relevant files are not included in the MiNT distribution.
# You might as well use the modified Unix Makefile (read the comments at the
# beginning of Makefile how to configure it).
#
# This Makefile is loosely based on the GNU Makefile conventions found in
# standards.info.
#
# Compiling Vim, summary:
#
#	3. make
#	4. make install
#
# Compiling Vim, details:
#
# Edit this file for adjusting to your system. You should not need to edit any
# other file for machine specific things!
# The name of this file MUST be Makefile (note the uppercase 'M').
#
# 1. Edit this Makefile
#    The defaults for Vim should work on most machines, but you may want to
#    uncomment some lines or make other changes below to tune it to your
#    system, compiler or preferences.
#    - Uncomment one of CFLAGS and/or CC if you have something very special or
#      want to tune the optimizer.
#    - Search for the name of your system to see if it needs anything special.
#    - A few versions of make use '.include "file"' instead of 'include file'.
#      Adjust the include line below if yours does.
#
# 2. Edit feature.h
#    Only if you do not agree with the default compile features, e.g.:
#       - you want Vim to be as vi compatible as it can be
#       - you want to use Emacs tags files
#       - you want right-to-left editing (Hebrew)
#       - you want 'langmap' support (Greek)
#       - you want to remove features to make Vim smaller
#
# 3. "make"
#    Will first run ./configure with the options in this file. Then it will
#    start make again on this Makefile to do the compiling. You can also do
#    this in two steps with:
#		make config
#		make
#    The configuration phase creates/overwrites config.h and config.mk.
#    Configure is created with autoconf.  It can detect different features of
#    your system and act accordingly.  However, it is not correct for all
#    systems.
#    If you changed something do this to run configure again:
#		make distclean
#		make config
#		make
#
#    - If you do not trust the automatic configuration code, then inspect
#      config.h and config.mk, before starting the actual build phase. If
#      possible edit this Makefile, rather than config.mk -- especially look
#      the definitions of VIM_HLP and SYS_VIMRC_FILE below. Note that the
#      configure phase overwrites config.mk and config.h again.
#    - If you get error messages, find out what is wrong and try to correct it
#      in this Makefile. You may need to do "make distclean" when you
#      change anything that configure uses (e.g. switching from an old C
#      compiler to an ANSI C compiler). Only when ./configure does something
#      wrong you may need to change one of the other files. If you find a
#      clean way to fix the problem, consider sending a note to the author of
#      autoconf (bug-gnu-utils@prep.ai.mit.edu) or Vim (mool@oce.nl). Don't
#      bother to do that when you made a hack solution for a non-standard
#      system.
#
# 4. "make install"
#    If the new Vim seems to be working OK you can install it and the
#    documentation in the appropriate location. The default is "/usr/local".
#    Change "PREFIX" below to change the location. "pathdef.c" will be
#    compiled again after changing this to make the executable know where the
#    help files are located.
#    Note that any existing executable and documentation is removed or
#    overwritten.  If you want to keep it you will have to make a backup copy
#    first.

### This Makefile has been succesfully tested on these systems.
### Check the (*) column for remarks, listed below.
### Later code changes may cause small problems, otherwise Vim is supposed to
### compile and run without problems.

#system:              configurations:    	     version (*) tested by:
#-------------        ------------------------       -------  -  ----------
#MiNT 1.12.5          gcc gcc-2.6.1                     3.29 (M) Jens Felderhoff
#MiNT 1.12.6          gcc gcc-2.6.1       -GUI          4.6b (M) Jens Felderhoff
#MiNT 1.12.6          gcc gcc-2.6.1       -GUI          4.6  (M) Jens Felderhoff


# (*)  Remarks:
#
# (M)  See below for details.

#DO NOT CHANGE the next line, we need it for configure to find the compiler
#instead of using the default from the "make" program.
#Use a line further down to change the value for CC.
CC=

######################## config.mk ########################
# At this position config.mk is included. When starting from the
# distribution it is almost empty. After running ./configure it contains
# settings that have been discovered for your system. Settings below this
# include override settings in config.mk!

# (X) How to include config.mk depends on the version of "make" you have, if
#     the current choice doesn't work, try the other one.

include config.mk
#.include "config.mk"

# Include the configuration choices first, so we can override everything
# below. As shipped, this file contains a target that causes to run
# configure. Once configure was run, this file contains a list of
# make variables with predefined values instead. Thus any second invocation
# of make, will buid Vim.

# You can give a lot of options to configure.
# Change this to your desire and do 'make config' afterwards

### For creating a plain Vim without any X11 related fancies use this one:
### (otherwise Vim configure will try to include xterm titlebar access)
CONF_OPT_X = --without-x

### COMPILER -- the default from configure will mostly be fine, no need to
### change this, just an example. If a compiler is defined here, configure
### will use it rather than probing for one. It is dangerous to change this
### after configure was run. Make will use your choice then -- but beware:
### Many things may change with another compiler. It is wise to run 'make
### distclean' and start all over again.
#CC = cc
#CC = gcc

###
### COMPILER FLAGS - change as you please. Either before running configure
### or afterwards. For examples see below.
### Linux: When using -g, you will probably get a statically linked
### executable!!
### When not defined, configure will try to use -O2 -g for gcc and -O for cc.
#CFLAGS = -g
#CFLAGS = -O
#CFLAGS = -g -O2 -fno-strength-reduce -Wall -Wshadow -Wmissing-prototypes

### Often used for GCC: optimizing, lot of optimizing, debugging
#CFLAGS = -O2 -fno-strength-reduce -Wall -Wshadow -Wmissing-prototypes
#CFLAGS = -O6 -fno-strength-reduce -Wall -Wshadow -Wmissing-prototypes
#CFLAGS = -g -DDEBUG -Wall -Wshadow -Wmissing-prototypes

#####################################################
###  Specific systems, check if yours is listed!  ###
#####################################################

### Uncomment things here only if the values chosen by configure are wrong.
### It's better to adjust configure.in and run autoconf, if you can!
### Then send the required changes to configure.in to the bugs list.

### (M)  MiNT with gcc 2.6.1 and gdb 3.5
CC = gcc -mint
CFLAGS = -g -O
POSTPROCESS = fixstk 20k $(VIMTARGET)
DBGLDFLAGS = -B/gnu/bin/sym-
DBGTARGET = $(VIMTARGET).sym

### If you want to use ncurses library instead of the automatically found one
### after changing this, you need to do "make distclean; make".
#TERM_LIB = -lncurses

##################### end of system specific lines ###################

### Name of target(s)
VIMTARGET	= vim
EXTARGET	= ex
VIEWTARGET	= view
GVIMTARGET	= g$(VIMTARGET)
GEXTARGET	= g$(EXTARGET)
GVIEWTARGET	= g$(VIEWTARGET)
RVIMTARGET	= r$(VIMTARGET)
REXTARGET	= r$(EXTARGET)
RVIEWTARGET	= r$(VIEWTARGET)
RGVIMTARGET	= r$(GVIMTARGET)
RGEXTARGET	= r$(GEXTARGET)
RGVIEWTARGET	= r$(GVIEWTARGET)

#DBGTARGET = $(VIMTARGET)

### Names of the tools that are also made
TOOLS = ctags/ctags xxd/xxd

### Installation directory.  The defaults come form configre.
### EXEC_PREFIX is for the executable   (default "/usr/local")
### PREFIX      is for the manual pages (default "/usr/local")
### DATADIR     is for the other files  (default "/usr/local/lib" or
###						"/usr/local/share")
### They may be different when using different architectures for the
### executable and a common directory for the other files.
###
### Uncomment the next three lines to install Vim in your home directory.
#PREFIX      = $(HOME)
#EXEC_PREFIX = $(HOME)
#DATADIR     = $(HOME)

### Uncomment the next line only if the value of PREFIX is different from
### EXEC_PREFIX
#PREFIX_MKDIR = $(PREFIX)

### Uncomment the next line only if the value of DATADIR is different from
### EXEC_PREFIX
#DATADIR_MKDIR = $(DATADIR)

### Location of executable
BINLOC = $(EXEC_PREFIX)/bin

### Location of man page
MANLOC = $(PREFIX)/man
MANDIR = /man1

### Location of Vim files (should not need to be changed)
VIMDIR = /vim
HELPSUBDIR = /doc
SYNSUBDIR = /syntax

### VIMLOC      common root of the Vim files
### VIMRCLOC    location for system-wide vimrc and gvimrc files
### HELPSUBLOC  location for help files
### SYNSUBLOC   location for syntax files
### MENULOC	location for menu file
### You can override these if you want them somewhere else (e.g. Debian wants
### VIMRCLOC to be /etc).
VIMLOC     = $(DATADIR)$(VIMDIR)
VIMRCLOC   = $(VIMLOC)
HELPSUBLOC = $(VIMLOC)$(HELPSUBDIR)
MENULOC    = $(VIMLOC)
SYNSUBLOC  = $(VIMLOC)$(SYNSUBDIR)

### These go into pathdef.c
VIM_HLP         = $(HELPENDLOC)/help.txt
SYS_VIMRC_FILE  = $(VIMRCLOC)/vimrc
SYS_GVIMRC_FILE = $(VIMRCLOC)/gvimrc
SYS_MENU_FILE   = $(MENULOC)/menu.vim

### Program to run on installed binary
STRIP = strip

### Permissions for vim binary
BINMOD = 755

### Permissions for man page
MANMOD = 644

### Permissions for help files
HELPMOD = 644

### Permission for menu file
MENUMOD = 644

### Permissions for bin/man/help directories that are created
DIRMOD = 755

# Where to copy the man and help files from
HELPSOURCE = ../doc

# Where to copy the menu file from
MENUSOURCE = ..

# Where to copy the syntax files from
SYNSOURCE = ../syntax

# If you are using Linux, you might want to use this to make vim the
# default vi editor, it will create a link to Vim when doing "make install".
#LINKIT = -ln -f -s /usr/bin/vi $(BINLOC)/$(VIMTARGET)

### our grand parent directory should know who we are...
### only used for "make tar"
VIMVERSION = `eval "basename \`cd ../../; pwd\`"`

### Command to create dependencies based on #include "..."
### prototype headers are ignored due to -DPROTO, system
### headers #include <...> are ignored if we use the -MM option, as
### e.g. provided by gcc-cpp.
CPP_DEPEND = $(CC) -M$(CPP_MM) -DPROTO $(DEFS) $(EXTRA_DEFS)


################################################
##   no changes required below this line      ##
################################################

SHELL = /bin/sh

.SUFFIXES:
.SUFFIXES: .c .o .pro

ALL_CFLAGS = $(DEFS) $(EXTRA_DEFS) $(CFLAGS) $(PERL_CFLAGS)

ALL_LIBS = $(LIBS) $(EXTRA_LIBS)

PFLAGS = $(PROTO_FLAGS) -DPROTO $(DEFS) $(EXTRA_DEFS) $(PERL_CFLAGS)

#     BASIC_INCL and BASIC_SRC: files that are always used
#
#                 INCL and SRC: files used for current configuration
#                     ALL_INCL: files used for make depend and ctags
#                      ALL_SRC: files used for make depend

BASIC_INCL = ascii.h ex_docmd.h ex_cmds.h config.h feature.h globals.h \
		keymap.h macros.h option.h osdef.h proto.h regexp.h \
		structs.h term.h os_unix.h os_unixx.h version.h vim.h

INCL =  $(BASIC_INCL)

ALL_INCL = $(BASIC_INCL) if_perl.xs if_python.c

BASIC_SRC = \
	buffer.c \
	charset.c \
	digraph.c \
	edit.c \
	eval.c \
	ex_cmds.c \
	ex_docmd.c \
	ex_getln.c \
	fileio.c \
	getchar.c \
	main.c \
	mark.c \
	memfile.c \
	memline.c \
	message.c \
	misc1.c \
	misc2.c \
	normal.c \
	ops.c \
	option.c \
	os_unix.c \
	pathdef.c \
	quickfix.c \
	regexp.c \
	screen.c \
	search.c \
	syntax.c \
	tag.c \
	term.c \
	ui.c \
	undo.c \
	version.c \
	window.c

SRC =	$(BASIC_SRC)

ALL_SRC = $(BASIC_SRC)

OBJ = \
	buffer.o \
	charset.o \
	digraph.o \
	edit.o \
	eval.o \
	ex_cmds.o \
	ex_docmd.o \
	ex_getln.o \
	fileio.o \
	getchar.o \
	main.o \
	mark.o \
	memfile.o \
	memline.o \
	message.o \
	misc1.o \
	misc2.o \
	normal.o \
	ops.o \
	option.o \
	os_unix.o \
	pathdef.o \
	quickfix.o \
	regexp.o \
	screen.o \
	search.o \
	syntax.o \
	tag.o \
	term.o \
	ui.o \
	undo.o \
	window.o \
	$(PERL_OBJ)

# Default target is making the executable
all: $(VIMTARGET) $(TOOLS)

# Run ./configure with all the setting above.
config: configure
	CC="$(CC)" CFLAGS="$(CFLAGS)" TERM_LIB="$(TERM_LIB)" \
	$(CONF_SHELL) ./configure $(CONF_OPT_PERL)

# When configure.in has changed, run autoconf to produce configure
# If you don't have autoconf, use the configure that's there
configure: configure.in
	autoconf
# Fix a bug in autoconf 2.8 and 2.9 (not needed for 2.10, which we use now):
#	mv -f configure configure.wrong
#	sed -e 's+$$ac_cv_lib_dnet+$$ac_cv_lib_$$ac_lib_var+' configure.wrong > configure
#	rm -f configure.wrong
	chmod 755 configure

# Re-execute this Makefile to include the new config.mk produced by configure
# Only used when typing "make" with a fresh config.mk.
myself:
	$(MAKE) -f Makefile all


# Link the target for normal use
$(VIMTARGET): $(OBJ) version.c version.h
	$(CC) -c $(ALL_CFLAGS) version.c
	$(CC) $(LDFLAGS) -o $(VIMTARGET) $(OBJ) version.o $(ALL_LIBS)
	$(POSTPROCESS)

# Link the target for debugging
debug: $(OBJ) version.c version.h
	$(CC) -c $(ALL_CFLAGS) version.c
	$(CC) $(LDFLAGS) $(DBGLDFLAGS) -o $(DBGTARGET) -g $(OBJ) version.o $(ALL_LIBS)

ctags/ctags: ctags/Makefile
	cd ctags; $(MAKE) -f Makefile

ctags/Makefile: ctags/Makefile.in
	cd ctags; ./configure

xxd/xxd: xxd/xxd.c
	cd xxd; CC="$(CC)" CFLAGS="$(CFLAGS)" $(MAKE) -f Makefile.unix

notags:
	rm -f tags

# Note: tags is made for the currently configured version.
ctags tags TAGS: notags
	$(CTAGS) $(SRC) $(ALL_INCL)

#
# Execute the test scripts.  Run these after compiling Vim, before installing.
#
# This will produce a lot of garbage on your screen, including a few error
# messages.  Don't worry about that.
# If there is a real error, there will be a difference between "test.out" and
# a "test99.ok" file.
# If everything is allright, the final message will be "ALL DONE".
#
test:
	cd testdir; $(MAKE) -f Makefile $(TESTTARGET)

testclean:
	-rm -f testdir/*.out

#
# Avoid overwriting an existing executable, somebody might be running it and
# overwriting it could cause it to crash.  Deleting it is OK, it won't be
# really deleted until all running processes for it have exited.  It is
# renamed first, in case the deleting doesn't work.
#
# If you want to keep an older version, rename it before running "make
# install".
#
install: installvim installinks

installvim: $(VIMTARGET) $(TOOLS) $(EXEC_PREFIX) $(BINLOC) $(PREFIX) $(MANLOC) \
		$(MANLOC)$(MANDIR) $(VIMLOC) $(HELPSUBLOC) $(SYNSUBLOC)
	-mv -f $(BINLOC)/$(VIMTARGET) $(BINLOC)/$(VIMTARGET).old
	-rm -f $(BINLOC)/$(VIMTARGET).old
	cp $(VIMTARGET) $(BINLOC)
	$(STRIP) $(BINLOC)/$(VIMTARGET)
	chmod $(BINMOD) $(BINLOC)/$(VIMTARGET)
# may create a link to the new executable from /usr/bin/vi
	-$(LINKIT)
# install the help files; first adjust the contents for the location
	sed -e s+/usr/local/lib/vim+$(HELPENDLOC)+ $(HELPSOURCE)/vim.1 > $(MANLOC)$(MANDIR)/vim.1
	chmod $(MANMOD) $(MANLOC)$(MANDIR)/vim.1
	cp $(HELPSOURCE)/*.txt $(HELPSUBLOC)
	cp $(HELPSOURCE)/tags $(HELPSUBLOC)
	chmod $(HELPMOD) $(HELPSUBLOC)/*.txt $(HELPSUBLOC)/tags
# install the menu file
	cp $(MENUSOURCE)/menu.vim $(SYS_MENU_FILE)
	chmod $(MENUMOD) $(SYS_MENU_FILE)
# install the syntax files
	cp $(SYNSOURCE)/*.vim $(SYNSUBLOC)
	chmod $(HELPMOD) $(SYNSUBLOC)/*.vim
# install helper program ctags
	cd ctags; $(MAKE) -f Makefile install
# install helper program xxd
	-mv -f $(BINLOC)/xxd $(BINLOC)/xxd.old
	-rm -f $(BINLOC)/xxd.old
	cp xxd/xxd $(BINLOC)
	$(STRIP) $(BINLOC)/xxd
	chmod $(BINMOD) $(BINLOC)/xxd
	cp $(HELPSOURCE)/xxd.1 $(MANLOC)$(MANDIR)
	chmod $(MANMOD) $(MANLOC)$(MANDIR)/xxd.1

$(EXEC_PREFIX) $(BINLOC) $(PREFIX_MKDIR) $(DATADIR_MKDIR) $(MANLOC) $(MANLOC)$(MANDIR) $(VIMLOC) $(HELPSUBLOC) $(SYNSUBLOC):
	-ctags/mkinstalldirs $@
	-chmod $(DIRMOD) $@

# create links from various names to vim.  This is only done when the links
# (or executables with the same name) don't exist yet.
installinks: $(BINLOC)/$(EXTARGET) $(BINLOC)/$(VIEWTARGET) $(BINLOC)/$(RVIMTARGET) $(BINLOC)/$(REXTARGET) $(BINLOC)/$(RVIEWTARGET)

$(BINLOC)/$(EXTARGET):
	cd $(BINLOC); ln -s $(VIMTARGET) $(EXTARGET)

$(BINLOC)/$(VIEWTARGET):
	cd $(BINLOC); ln -s $(VIMTARGET) $(VIEWTARGET)

$(BINLOC)/$(RVIMTARGET):
	cd $(BINLOC); ln -s $(VIMTARGET) $(RVIMTARGET)

$(BINLOC)/$(REXTARGET):
	cd $(BINLOC); ln -s $(VIMTARGET) $(REXTARGET)

$(BINLOC)/$(RVIEWTARGET):
	cd $(BINLOC); ln -s $(VIMTARGET) $(RVIEWTARGET)

uninstall:
	rm -f $(BINLOC)/$(VIMTARGET) $(MANLOC)$(MANDIR)/vim.1
	rm -f $(HELPSUBLOC)/vim_*.txt $(HELPSUBLOC)/tags

# Clean up all the files that have been produced, except configure's
# We support common typing mistakes for Juergen! :-)
clean celan: testclean
	-rm -f *.o core vim.core $(VIMTARGET) xxd/*.o $(TOOLS) osdef.h 
	-rm -f pathdef.c mkcmdtab ex_cmds.h
	-rm -f conftest* *~ link.sed
	-cd ctags; $(MAKE) -f Makefile clean

shadow:	doc menu.vim syntax
	mkdir shadow
	cd shadow; ln -s ../*.[ch] ../*.in ../*.sh ../*.xs ../toolcheck ../proto ../*.tab ../configure .
	cd shadow; rm -f osdef.h config.h ex_cmds.h pathdef.c link.sed
	cp Makefile shadow
	echo "the first targets to make vim are: scratch config myself" > shadow/config.mk
# TODO: how do we make a nice shadow for ctags?
	cd shadow; ln -s ../ctags .
	mkdir shadow/xxd
	cd shadow/xxd; ln -s ../../xxd/*.[ch] ../../xxd/Make* .
	mkdir shadow/testdir
	cd shadow/testdir; ln -s ../../testdir/Makefile \
				 ../../testdir/*.in  ../../testdir/*.ok .

# Links needed for doing "make install" in a shadow directory.
doc:
	ln -s ../doc .
menu.vim:
	ln -s ../menu.vim .

syntax:
	ln -s ../syntax .


# Start configure from scratch
scrub scratch:
	rm -f config.status config.cache config.h config.log

distclean: scratch clean
	echo "the first targets to make vim are: scratch config myself" > config.mk

tar: clean
	echo packing $(VIMVERSION) ...
	VIMVERSION=$(VIMVERSION); cd ../../..; set -x; \
	  tar cvf $$VIMVERSION.tar $$VIMVERSION; \
	  gzip -nf $$VIMVERSION.tar || gzip -f $$VIMVERSION.tar

dist: distclean tar

check:

mdepend:
	-@rm -f Makefile~
	cp Makefile Makefile~
	sed -e '/\#\#\# Dependencies/q' < Makefile > tmp_make
	@for i in $(ALL_SRC) ; do \
	  echo "$$i" ; \
	  echo `echo "$$i" | sed -e 's/.c$$/.o/'`": $$i" `\
            $(CPP) $$i |\
            grep '^# .*"\./.*\.h"' |\
            sort -t'"' -u +1 -2 |\
            sed -e 's/.*"\.\/\(.*\)".*/\1/'\
          ` >> tmp_make ; \
	done
	mv tmp_make Makefile

depend:
	-@rm -f Makefile~
	cp Makefile Makefile~
	sed -e '/\#\#\# Dependencies/q' < Makefile > tmp_make
	-for i in $(ALL_SRC); do echo $$i; \
			$(CPP_DEPEND) $$i >> tmp_make; done
	mv tmp_make Makefile

###########################################################################

# Used when .o files are in src directory
.c.o:
	$(CC) -c -I$(srcdir) $(ALL_CFLAGS) $<

# Used when .o files are in src/objects directory
#$(OBJ): $$(*F).c
#	$(CC) -c -I$(srcdir) $(ALL_CFLAGS) $(*F).c -o $@

if_perl.c: if_perl.xs
	$(PERL) $(PERLLIB)/ExtUtils/xsubpp -typemap \
	    $(PERLLIB)/ExtUtils/typemap if_perl.xs > $@

osdef.h: osdef.sh config.h osdef1.h.in osdef2.h.in
	CC="$(CC) $(ALL_CFLAGS)" srcdir=${srcdir} sh $(srcdir)/osdef.sh

pathdef.c: Makefile config.mk
	-@echo creating pathdef.c
	-@echo '/* pathdef.c */' > pathdef.c
	-@echo '/* This file is automatically created by Makefile' >> pathdef.c
	-@echo ' * DO NOT EDIT!  Change Makefile only. */' >> pathdef.c
	-@echo '#include "vim.h"' >> pathdef.c
	-@echo 'char_u *default_vim_dir = (char_u *)"$(VIMLOC)";' >> pathdef.c
	-@echo 'char_u *all_cflags = (char_u *)"$(CC) -c -I$(srcdir) $(ALL_CFLAGS)";' >> pathdef.c
	-@echo 'char_u *all_lflags = (char_u *)"$(CC) $(LDFLAGS) -o $(VIMTARGET) $(ALL_LIBS)";' >> pathdef.c
	-@sh $(srcdir)/pathdef.sh

Makefile:
	@echo The name of the makefile MUST be "Makefile" (with capital M)!!!!

ex_cmds.h: ex_docmd.h mkcmdtab
	./mkcmdtab ex_docmd.h ex_cmds.h

mkcmdtab: mkcmdtab.c osdef.h
	$(CC) -I$(srcdir) $(ALL_CFLAGS) -o mkcmdtab mkcmdtab.c

###############################################################################
### (automatically generated by 'make depend')
### Dependencies:
buffer.o: buffer.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
charset.o: charset.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
ex_cmds.o: ex_cmds.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h ex_cmds.h
cmdline.o: cmdline.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
digraph.o: digraph.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
edit.o: edit.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
fileio.o: fileio.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
getchar.o: getchar.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
misc2.o: misc2.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h
main.o: main.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
mark.o: mark.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
memfile.o: memfile.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
memline.o: memline.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
message.o: message.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
misc1.o: misc1.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
normal.o: normal.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
ops.o: ops.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h term.h \
  macros.h structs.h globals.h proto.h regexp.h option.h
option.o: option.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
pathdef.o: pathdef.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h
quickfix.o: quickfix.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h option.h
regexp.o: regexp.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
screen.o: screen.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
search.o: search.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
tag.o: tag.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h term.h \
  macros.h structs.h globals.h proto.h regexp.h option.h
term.o: term.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h option.h proto.h regexp.h
ui.o: ui.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
undo.o: undo.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
os_unix.o: os_unix.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h option.h proto.h regexp.h os_unixx.h
version.o: version.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h \
  keymap.h term.h macros.h structs.h globals.h proto.h regexp.h
window.o: window.c vim.h config.h feature.h os_unix.h osdef.h os_mint.h ascii.h keymap.h \
  term.h macros.h structs.h globals.h proto.h regexp.h option.h
