" Vim syntax file
" Language:	Fortran
" Maintainer:	Mario Eusebio <bio@dq.fct.unl.pt>
" Last change:	1997 April 20

" Remove any old syntax stuff hanging around
syntax clear

syntax case ignore

syntax keyword fortranStatement	return continue 
syntax keyword fortranLabel		go goto 
syntax keyword fortranConditional	if else elseif 
syntax keyword fortranConditional	then 
syntax keyword fortranRepeat		do 

syntax keyword fortranTodo contained	TODO

" String and Character contstants
" Highlight special characters (those which have a backslash) differently
" This is also valid with Gnu Fortran
syntax match   fortranSpecial contained "\\[0-9][0-9][0-9]\|\\."
syntax region  fortranString	  start=+"+  skip=+\\\\\|\\"+  end=+"+  contains fortranSpecial
syntax region  fortranString	  start=+'+  skip=+\\\\\|\\"+  end=+'+  contains fortranSpecial

"syntax match  fortranIdentifier	"\<[a-zA-Z_][a-zA-Z0-9_]*\>"

"syntax match  fortranDelimiter	"[()]"

syntax match  fortranLabelNumber          "^[0-9]\=[0-9]\=[0-9]\=[0-9]\=[0-9]\>"
syntax match  fortranLabelNumber          "^\ [0-9]\=[0-9]\=[0-9]\=[0-9]\>"
syntax match  fortranLabelNumber          "^\ \ [0-9]\=[0-9]\=[0-9]\>"
syntax match  fortranLabelNumber          "^\ \ \ [0-9]\=[0-9]\>"
syntax match  fortranLabelNumber          "^\ \ \ \ [0-9]\>"

syntax match  fortranNoLabelNumber "^[0-9][0-9][0-9][0-9][0-9][^\ ^\t].*"
syntax match  fortranNoLabelNumber          "^\ [0-9][0-9][0-9][0-9][^\ ^\t].*"
syntax match  fortranNoLabelNumber          "^\ \ [0-9][0-9][0-9][^\ ^\t].*"
syntax match  fortranNoLabelNumber          "^\ \ \ [0-9][0-9][^\ ^\t].*"
syntax match  fortranNoLabelNumber          "^\ \ \ \ [0-9][^\ ^\t].*"

" I don't like fortranPoint* form of numbers
syntax match  fortranDoubleNumber     "-\=\<[0-9]\+\.[0-9]\+[Dd]-\=[0-9]\+\>"
"syntax match  fortranPointDoubleNumber     "-\=\<[0-9]\+\.[Dd]-\=[0-9]\+\>"
"syntax match  fortranPointDoubleNumber     "-\=\.[0-9]\+[Dd]-\=[0-9]\+\>"
syntax match  fortranRealNumber       "-\=\<[0-9]\+\.[0-9]\+[eE]-\=[0-9]\+\>"
"syntax match  fortranPointRealNumber       "-\=\<[0-9]\+\.[eE]-\=[0-9]\+\>"
"syntax match  fortranPointRealNumber       "-\=\.[0-9]\+[eE]-\=[0-9]\+\>"
syntax match  fortranRealNumber       "-\=\<[0-9]\+\.[0-9]\+\>"
"syntax match  fortranPointRealNumber       "-\=\<[0-9]\+\."
"syntax match  fortranPointRealNumber       "-\=\.[0-9]\+\>"
syntax match  fortranNumber           "-\=\<[0-9]\+\>"

syntax match  fortranComment		"^[cC*].*" contains fortranTodo
syntax match  fortranComment		"!.*" contains fortranTodo

syntax match fortranppPreCondit       "^#define"
syntax match fortranppPreCondit       "^#include"

" any character other than a c in the first column
" would be an error
" does not work with keywords 
syntax match  fortranCommentError	"^[^\ ^0-9^\t].*"
syntax match  fortranCommentError	"^\ [^\ ^0-9^\t].*"
syntax match  fortranCommentError	"^\ \ [^\ ^0-9^\t].*"
syntax match  fortranCommentError	"^\ \ \ [^\ ^0-9^\t].*"
syntax match  fortranCommentError	"^\ \ \ \ [^\ ^0-9^\t].*"

syntax match  fortranContinueLine  "^\ \ \ \ \ [^\t^\ ]"

" If you don't like initial tabs in fortran
"syntax match fortranShowIniTab "^\t.*"
" If you don't like tabs at all
"syntax match fortranShowTab "\t"

" You can substitute [.] for \.
syntax match fortranOperator       "[.]and[.]"
syntax match fortranOperator       "[.]eq[.]"
syntax match fortranOperator       "[.]eqv[.]"
syntax match fortranOperator       "[.]ge[.]"
syntax match fortranOperator       "[.]gt[.]"
syntax match fortranOperator       "[.]le[.]"
syntax match fortranOperator       "[.]lt[.]"
syntax match fortranOperator       "[.]ne[.]"
syntax match fortranOperator       "[.]neqv[.]"
syntax match fortranOperator       "[.]not[.]"
syntax match fortranOperator       "[.]or[.]"

syntax match fortranLogicalConstant       "[.]true[.]"
syntax match fortranLogicalConstant       "[.]false[.]"

syntax keyword fortranType			integer real real*8
syntax keyword fortranType			precision complex 
syntax keyword fortranType			character dimension 
syntax keyword fortranType			logical double 
syntax keyword fortranStructure		equivalence common 
syntax keyword fortranStructure		implicit parameter 
syntax keyword fortranStorageClass	program subroutine data 
syntax keyword fortranStorageClass	external intrinsic 
syntax keyword fortranStorageClass	function end 
syntax keyword fortranReadWrite       read write print 
syntax keyword fortranReadWrite       open close rewind 
syntax keyword fortranMemory       save 
syntax keyword fortranStopPause       stop pause 
syntax keyword fortranImplicit    abs acos atan asin cos 
syntax keyword fortranImplicit    cosh aimag anint atan2 
syntax keyword fortranImplicit    char cmplx conjg dble 
syntax keyword fortranImplicit    dim dprod exp ichar index 
syntax keyword fortranImplicit    int len lge lgt lle llt 
syntax keyword fortranImplicit    log log10 max min mod 
syntax keyword fortranImplicit    sign sqrt tan 

syntax sync ccomment fortranComment

if !exists("did_fortran_syntax_inits")
  let did_fortran_syntax_inits = 1
  " The default methods for highlighting.  Can be overridden later
  highlight link fortranLabel		Identifier
  highlight link fortranLabelNumber		Identifier
  highlight link fortranNoLabelNumber		Error
  highlight link fortranConditional	fortranStatement
  highlight link fortranRepeat	fortranStatement
  highlight link fortranDoubleNumber	fortranNumber
  highlight link fortranRealNumber	fortranNumber
  highlight link fortranNumber	Number
  highlight link fortranCommentError	Error
  highlight link fortranOperator	fortranStatement
  highlight link fortranLogicalConstant	fortranStatement
  highlight link fortranStructure	fortranType
  highlight link fortranStorageClass	fortranType
  highlight link fortranContinueLine	Statement
  highlight link fortranStatement	Statement
  highlight link fortranppPreCondit	PreCondit
  highlight link fortranType		Type
  highlight link fortranString	String
  highlight link fortranComment	Comment
  highlight link fortranSpecial	Special
  highlight link fortranTodo		Todo
  highlight link fortranMemory     Identifier
  highlight link fortranStopPause     Identifier
  highlight link fortranImplicit     Identifier

  " optional highlighting
  "highlight link fortranppPreCondit   PreCondit
  "highlight link fortranPointRealNumber    fortranString
  "highlight link fortranPointDoubleNumber  fortranString
  "highlight link fortranShowIniTab         Error
  "highlight link fortranShowTab         Error
  "highlight link fortranDelimiter     Identifier

  "highlight link fortranIdentifier	Identifier
endif
